/*******************************************************************************
 * Project: SLIM5 experiment - Test Beam Reconstruction Software
 * Package: Sbt
 *    File: $Id$
 * Authors:
 *   Marco Bomben                 Marco.Bomben@ts.infn.it, INFN-Trieste
 * History:
 *    - Dec 22: creation
 ******************************************************************************/

// -- CLASS DESCRIPTION [MAIN] --
// the implementation of the SbtGrandFit class

#include <sstream>
#include "Sbt/SbtGrandFit.hh"

SbtGrandFit::SbtGrandFit( std::vector<SbtTrack*> listOfTracks ) : _listOfTracks(listOfTracks) {
  _ntracks = _listOfTracks.size();

  if ( _ntracks < 10 ) {
    std::cout << "Unable to fit all params with only: " << _ntracks << " tracks.\n";
    std::cout << "Exiting now...\n";
    exit( 1 );
  }

  RemoveBadTracks();
  _nparms = 4 + _ntracks*4; // 4 kind of errors + 4 params per track
  // Initialize TMinuit via generic fitter interface
  _minuit = TVirtualFitter::Fitter(0, _nparms);

  DeclareParams();


}

void SbtGrandFit::RemoveBadTracks() {

  // loop over tracks to retain only good tracks (i.e. with 4 space-points )

  std::vector<SbtTrack*>::iterator aTrack;

  for ( aTrack = _listOfTracks.begin(); aTrack != _listOfTracks.end(); aTrack++ ) {

    if ( ( (*aTrack)->GetSpacePointList() ).size() != 4) {
      _listOfTracks.erase(aTrack);
    }
    

  }
  _ntracks =  _listOfTracks.size();

  if ( _ntracks < 10 ) {
    std::cout << "Unable to fit all params with only " << _ntracks << " good trakcs.\n";
    std::cout << "Exiting now...\n";
    exit( 2 );
  }

}

void SbtGrandFit::DeclareParams() {

  if ( _minuit ) {
    _minuit->SetFCN(calcChi2);

    _minuit->SetParameter(0, "pSide_CS1",     7.2e-4,  1.e-5, 0.,1e-2);
    _minuit->SetParameter(1, "pSide_CS2",     6.0e-4,  1.e-5, 0.,1e-2);
    _minuit->SetParameter(2, "nSide_CS1",    14.4e-4,  1.e-5, 0.,1e-2);
    _minuit->SetParameter(3, "nSide_CS2",    10.0e-4,  1.e-5, 0.,1e-2);

    std::stringstream name; // stream for parameter name

    // loop over number of tracks; for each we define 4 parameters
    for ( Int_t ipar = 4; ipar < (_ntracks+1)*4  ; ipar++ ) {  

       Int_t iTrack = (ipar - 4)/4;


       /* x projection  */
  
       // qx
       name.str("");
       name << "qx of track " << setw(4) << iTrack;
  
       _minuit->SetParameter( ipar, (name.str()).c_str(), 0., 0.5, -1., 1.);
       
       // mx
       name.str("");
       name << "mx of track " << setw(4) << iTrack;
  
       _minuit->SetParameter( ipar+1, (name.str()).c_str(), 0., 0.05, -0.1, 0.1);
  
       /* y projection */
  
       // qy
       name.str("");
       name << "qy of track " << setw(4) << iTrack;
  
       _minuit->SetParameter( ipar+2, (name.str()).c_str(), 0., 0.5, -1., 1.);
       
       // my
       name.str("");
       name << "my of track " << setw(4) << iTrack;
  
       _minuit->SetParameter( ipar+3, (name.str()).c_str(), 0., 0.05, -0.1, 0.1);
  
    }

  } else {

    std::cout << "Fatal: fitter not defined: " << _minuit << "\n";
    std::cout << "Exiting now...\n";
    exit( 3 );
  }

  PrintParams();
  
}

void SbtGrandFit::calcChi2(Int_t &npar, Double_t *gin, Double_t &f, Double_t *x, Int_t iflag) {


  double globalChi2 = 0.;
  double chi2X(0.), chi2Y(0.);
  double trackChi2(0.);
  Double_t pSideErr_CS1 = x[0];
  Double_t pSideErr_CS2 = x[1];
  Double_t nSideErr_CS1 = x[2];
  Double_t nSideErr_CS2 = x[3];

  std::vector<SbtTrack*> listOfTracks = GetListOfTracks();

  std::vector<SbtTrack*>::iterator aTrack;

  Int_t iTrack = 0;

  std::cout << "Ready to ready _listOfTracks.size()...\n";
  std::cout << "ntracks: " << _listOfTracks.size() << "\n";

  /* Loop on all tracks to build the chisquare */
  for ( aTrack = listOfTracks.begin(); aTrack != listOfTracks.end(); aTrack++ ) {
    std::cout << "Looping on track number: " << iTrack << "\n"; 


    // Matrix of derivatives of line functional form wrt parameters
    TMatrixD Fx(4,2);
    // Matrix of derivatives of line functional form wrt parameters
    TMatrixD Fy(4,2);
    TMatrixD Mx(4,1);     // Vector of x SpacePoints measurements
    TMatrixD My(4,1);     // Vector of y SpacePoints measurements
    TMatrixD Vx(4,4);   // Vector of x SpacePoints measurements Errors
    TMatrixD Vy(4,4);   // Vector of y SpacePoints measurements Errors
    //initialize Vector Error Matrix to zero
    std::cout << "Zeroing error matrix...\n";
    for (int i=0; i<4; i++) {
      for (int j=0; j<4; j++) {
        Vx[i][j]=0.; Vy[i][j]=0.;
      }
    }
  
    TMatrixD lineParX(2,1); 
    TMatrixD lineParY(2,1); 

    std::cout << "Instruction for fitting...\n";

    Int_t parIndex = 4+iTrack*4;
    lineParX[0][1] = x[parIndex];
    lineParX[1][1] = x[parIndex+1];
    lineParY[0][1] = x[parIndex+2];
    lineParY[1][1] = x[parIndex+3];


    std::vector<SbtSpacePoint*> listOfSpacePoints =
                                (*aTrack)->GetSpacePointList();

    std::cout << "Ready to ready listOfSpacePoints.size()...\n";
    std::cout << "number of space points: " << listOfSpacePoints.size() << "\n";
    if ( listOfSpacePoints.size() != 4 ) continue;

    std::vector<SbtSpacePoint*>::iterator aSP;

    Int_t iSP = 0;

    /* Loop on SPs */
    for ( aSP=listOfSpacePoints.begin(); aSP!=listOfSpacePoints.end(); aSP++) {
      std::cout << "Looping on track number: " << iTrack << "\n"; 
      std::cout << "\tin particular SPpoint number: " << iSP << "\n";

      Fx[iSP][0] = 1.; Fx[iSP][1] = (*aSP)->GetZPosition();
      Fy[iSP][0] = 1.; Fy[iSP][1] = (*aSP)->GetZPosition();

      Mx[iSP][0] = (*aSP)->GetXPosition();
      My[iSP][0] = (*aSP)->GetYPosition();

      SbtHit* uHit = (*aSP)->GetHitU();
      SbtHit* vHit = (*aSP)->GetHitV();
      SbtCluster* uCluster = uHit->GetCluster();
      SbtCluster* vCluster = vHit->GetCluster();
      int uLength = uCluster->GetLength();
      int vLength = vCluster->GetLength();

      if( uLength == 1 ) Vx[iSP][iSP] = pSideErr_CS1;
      else Vx[iSP][iSP] = pSideErr_CS2;

      if( vLength == 1 ) Vy[iSP][iSP] = nSideErr_CS1;
      else Vy[iSP][iSP] = nSideErr_CS2;
  
      std::cout << "Fx: \n";
      Fx.Print();
      std::cout << "Fy: \n";
      Fy.Print();
      std::cout << "Vx: \n";
      Vx.Print();
      std::cout << "Vy: \n";
      Vy.Print();
  

      iSP++;

    } /* End Loop on SPs */
    std::cout << "Fx: \n";
    Fx.Print();
    std::cout << "Fy: \n";
    Fy.Print();
    std::cout << "Vx: \n";
    Vx.Print();
    std::cout << "Vy: \n";
    Vy.Print();

    TMatrixD estimateX = Fx*lineParX;
    TMatrixD estimateY = Fy*lineParY;

    TMatrixD DeltaX = Mx-estimateX;
    TMatrixD DeltaY = My-estimateY;

    TMatrixD VxInv = Vx; VxInv.Invert(); //define inverse x Error Matrix
    TMatrixD VyInv = Vy; VyInv.Invert(); //define inverse y Error Matrix
    TMatrixD DeltaXT = DeltaX;
    TMatrixD DeltaYT = DeltaY;
  


    chi2X = (DeltaXT*VxInv*DeltaX)[0][0];
    chi2Y = (DeltaYT*VyInv*DeltaY)[0][0];
    trackChi2 = chi2X + chi2Y;
    globalChi2 += trackChi2;

    iTrack++;
   

  } /* End Loop on tracks */  
  
  f = globalChi2;



 


}

void SbtGrandFit::PrintParams() {
  _arglist[0] = 0;
  _minuit->ExecuteCommand("SET PRINT",_arglist, 1);
}

int SbtGrandFit::Fit() {
  int status = -1;

  CallFcn();

  status =  CallMigrad();

  if ( status ) exit( status );

  status =  CallMinos();

  if ( status ) exit( status );


  return status;
}

void SbtGrandFit::CallFcn() {
  _arglist[0] = 1;
  _minuit->ExecuteCommand("CALL FCN", _arglist, 1);
}

int SbtGrandFit::CallMigrad() {
  int status = -1;
  _arglist[0] = 0;
  status = _minuit->ExecuteCommand("MIGRAD",_arglist, 0);
  return status;
}

int SbtGrandFit::CallMinos() {
  int status = -1;
  _arglist[0] = 0;
  status = _minuit->ExecuteCommand("MINOS", _arglist, 0);
  return status;
}
